# vile:confmode
# Configuration file for the color ls utility
# This file goes in the /etc directory, and must be world readable.
# You can copy this file to .dir_colors in your $HOME directory to override
# the system defaults.

# COLOR needs one of these arguments:
# 'tty' colorizes output to ttys, but not pipes (--color=auto),
# 'all' adds color characters to all output (--color=always),
# 'none' shuts colorization off (--color=never).
COLOR tty

# Extra command line options for ls go here.
# Basically these ones are:
#  -F = show '/' for dirs, '*' for executables, etc.
#  -T 0 = don't trust tab spacing when formatting ls output.
# -b = better support for special characters
OPTIONS -F -b -T 0

# Below are the color init strings for the basic file types. A color init
# string consists of one or more of the following numeric codes:
# Attribute codes:
# 00=none 01=bold 04=underscore 05=blink 07=reverse 08=concealed
# Text color codes:
# 30=black 31=red 32=green 33=yellow 34=blue 35=magenta 36=cyan 37=white
# Background color codes:
# 40=black 41=red 42=green 43=yellow 44=blue 45=magenta 46=cyan 47=white
NORMAL 00	# global default, although everything should be something.
FILE 00 	# normal file
DIR 01;34 	# directory
LINK 01;36 	# symbolic link
FIFO 40;33	# pipe
SOCK 01;35	# socket
BLK 40;33;01	# block device driver
CHR 40;33;01 	# character device driver
ORPHAN 40;31;01 # symlink to nonexistent file

# This is for files with execute permission:
EXEC 01;32

# List any file extensions like '.gz' or '.tar' that you would like ls
# to colorize below. Put the extension, a space, and the color init string.
# (and any comments you want to add after a '#')

# executables (bright green)
.cmd	01;32
.exe	01;32
.com	01;32
.btm	01;32
.bat	01;32

# archives or compressed (bright red)
.7z	01;31
.Z	01;31
.ace	01;31
.alz	01;31
.arc	01;31
.arj	01;31
.bz	01;31
.bz2	01;31
.cab	01;31
.cpio	01;31
.deb	01;31
.dwm	01;31
.dz	01;31
.ear	01;31
.esd	01;31
.gz	01;31
.jar	01;31
.lha	01;31
.lrz	01;31
.lz	01;31
.lz4	01;31
.lzh	01;31
.lzma	01;31
.lzo	01;31
.rar	01;31
.rpm	01;31
.rz	01;31
.sar	01;31
.swm	01;31
.t7z	01;31
.tar	01;31
.taz	01;31
.tbz	01;31
.tbz2	01;31
.tgz	01;31
.tlz	01;31
.txz	01;31
.tz	01;31
.tzo	01;31
.tzst	01;31
.war	01;31
.wim	01;31
.xz	01;31
.z	01;31
.zip	01;31
.zoo	01;31
.zst	01;31

# image formats
.asf	01;35
.avi	01;35
.bmp	01;35
.cgm	01;35
.dl	01;35
.emf	01;35
.flc	01;35
.fli	01;35
.flv	01;35
.gif	01;35
.gl	01;35
.jpeg	01;35
.jpg	01;35
.m2v	01;35
.m4v	01;35
.mjpeg	01;35
.mjpg	01;35
.mkv	01;35
.mng	01;35
.mov	01;35
.mp4	01;35
.mp4v	01;35
.mpeg	01;35
.mpg	01;35
.nuv	01;35
.ogm	01;35
.ogv	01;35
.ogx	01;35
.pbm	01;35
.pcx	01;35
.pdf	01;35
.pgm	01;35
.png	01;35
.ppm	01;35
.qt	01;35
.rm	01;35
.rmvb	01;35
.svg	01;35
.svgz	01;35
.tga	01;35
.tif	01;35
.tiff	01;35
.vob	01;35
.webm	01;35
.wmv	01;35
.xbm	01;35
.xcf	01;35
.xpm	01;35
.xwd	01;35
.yuv	01;35

# audio formats
.aac	00;36
.au	00;36
.flac	00;36
.m4a	00;36
.mid	00;36
.midi	00;36
.mka	00;36
.mp3	00;36
.mpc	00;36
.oga	00;36
.ogg	00;36
.opus	00;36
.ra	00;36
.spx	00;36
.wav	00;36
.xspf	00;36
