	# PROCESS BEGUN
	# argv[0] = 'ded'
	# argv[1] = '-t/tmp'
	# argv[2] = '-ctest2.cmd'
	# argv[3] = '-ltest2.out'
1D	# path: CM_TOOLS/test
1k	# path: CM_TOOLS
9R	# path: CM_TOOLS
1Z	# path: CM_TOOLS
1W	# path: CM_TOOLS
1F	# path: CM_TOOLS/test
1k	# path: CM_TOOLS/src
2k	# path: CM_TOOLS/bin
1E	# chdir CM_TOOLS/bin
1p444p	# chmod 444 ded
1p555p	# chmod 555 ded
1j
1t	# type "ded.hlp" (text file)
1\s
1b
1q
1h	# type "CM_TOOLS/bin/ded.hlp" (text file)
1\s
1b
1q
1\s
1g	# before "GROUP"
g
	# after  "GROUP"
1u	# before "LOGNAME"
u
	# after  "LOGNAME"
1!\\#\sthe\sscreen\sshows\sded\sand\sded.hlp\safter\stemporarily\schanging\spermissions,\sand\safter\styping\sthe\shelp\sfile\n
	# execute # the screen shows ded and ded.hlp after temporarily changing permissions, and after typing the help file
	# ELAPSED TIME
3l
1%ls\s-l\s%F\n
	# execute ls -l CM_TOOLS/test
\n	# Hit <RETURN> to continue
	# ELAPSED TIME
1!\\#\sjust\slisted\sthe\sother\sdirectory\n
	# execute # just listed the other directory
	# ELAPSED TIME
1.	# execute # just listed the other directory
	# ELAPSED TIME
1%\n
3l
1D	# path: CM_TOOLS/bin
1\L	# path: CM_TOOLS/bin
1\U	# path: CM_TOOLS
1E	# chdir CM_TOOLS
1F	# chdir CM_TOOLS/bin
1D	# path: CM_TOOLS/bin
1\L	# path: CM_TOOLS/bin
1\L	# path: CM_TOOLS/bin
2J	# path: CM_TOOLS/src
1J	# path: CM_TOOLS/test
1\R	# path: CM_TOOLS/test
1K	# path: CM_TOOLS/src
1l	# path: CM_TOOLS/src
1E	# chdir CM_TOOLS/src
1st
1\s
2#	# "Makefile"
	# "RCS"
	# "README"
	# "dedrc"
	# "boxchars.c"
	# "ded.c"
	# "ded2s.c"
	# "dedblip.c"
	# "dedcmds.c"
	# "dedcolor.c"
	# "deddoit.c"
	# "deddump.c"
	# "dedfind.c"
	# "dedfree.c"
	# "dedline.c"
	# "dedmake.c"
	# "dedmsgs.c"
	# "dedname.c"
	# "dedread.c"
	# "dedring.c"
	# "dedscan.c"
	# "dedshow.c"
	# "dedsigs.c"
	# "dedsize.c"
	# "dedsort.c"
	# "dedtags.c"
	# "dedtype.c"
	# "deduniq.c"
	# "dedview.c"
	# "dedwait.c"
	# "dlog.c"
	# "ftree.c"
	# "history.c"
	# "inline.c"
	# "restat.c"
	# "showpath.c"
	# "sortset.c"
	# "ded.h"
	# "version.h"
1/ded.a\n
	# "ded.a"
1s+t
1_
2+	# "ded.a"
	# "ded.hlp"
1!\\#\sselected\sthe\sfiles\sof\swhich\sonly\sone\ssuffix\sexists\n
	# execute # selected the files of which only one suffix exists
	# ELAPSED TIME
1!\n
3l
1_
1st
1k
1b
99b
1t	# type "Makefile" (text file)
1q
1\R
1\L
1\L	# (dedmsg) already at left margin
1\U
1\s
1k
99f
1j
1j
1j
1j
1\R
1\R
1\R
1\L
1\L
1\L
1I
1!\\#\sshows\sinode\sin\sthis\sscreen\n
	# execute # shows inode in this screen
	# ELAPSED TIME
1F	# chdir CM_TOOLS/test
1*
1st
1B	# chdir CM_TOOLS/src
1I
99b
1+	# "Makefile"
1+	# "RCS"
1+	# "README"
1_
1/\\.h$\n
	# "ded.h"
1+	# "ded.h"
1+	# "version.h"
1+	# "ded.hlp"
1k
1%ls\s-l\s#\n
	# execute ls -l ded.h version.h ded.hlp
\n	# Hit <RETURN> to continue
	# ELAPSED TIME
1_
1X
1b
1b
1X
9b
1M
1X
99b
1+	# "Makefile"
1+	# "RCS"
1+	# "README"
1+	# "dedrc"
1\t
1H
1w
1+	# "Makefile"
1+	# "RCS"
1X
1_
1n	# "ded.h"
1%echo\sF=%F\n
	# execute echo F=CM_TOOLS/test
\n	# Hit <RETURN> to continue
	# ELAPSED TIME
1:\F\b\b\b\bB=%B\n
	# execute echo B=CM_TOOLS/bin
\n	# Hit <RETURN> to continue
	# ELAPSED TIME
1:\F\b\b\b\bh=%h,\sr=%r,\se=%e\n
	# execute echo h=./, r=ded, e=.h
\n	# Hit <RETURN> to continue
	# ELAPSED TIME
1F	# chdir CM_TOOLS/test
1R
1\s
1F	# chdir CM_TOOLS
1D	# path: CM_TOOLS
2q	# path: CM_TOOLS/src
1F	# path: CM_TOOLS/test
1Q	# path: CM_TOOLS/src
1q	# chdir CM_TOOLS/src
1qyes\n
	# PROCESS ENDED
