	# PROCESS BEGUN
	# argv[0] = 'ded'
	# argv[1] = '-t/tmp'
	# argv[2] = '-ctest1.cmd'
	# argv[3] = '-ltest1.out'
1D	# path: CM_TOOLS/test
1k	# path: CM_TOOLS
9R	# path: CM_TOOLS
1Z	# path: CM_TOOLS
1F	# path: CM_TOOLS/test
1k	# path: CM_TOOLS/src
1k	# path: CM_TOOLS/certify
1k	# path: CM_TOOLS/bin
1E	# chdir CM_TOOLS/bin
1%ls\s-l\n
	# execute ls -l
\n	# Hit <RETURN> to continue
	# ELAPSED TIME
1st
1p44p	# chmod 445 ded
1p75p	# chmod 755 ded
1j
1t	# type "ded.hlp" (text file)
1q
1h	# type "CM_TOOLS/bin/ded.hlp" (text file)
1q
1D	# path: CM_TOOLS/bin
1j	# path: CM_TOOLS/certify
1j	# path: CM_TOOLS/src
1E	# chdir CM_TOOLS/src
1\s
1st
1Z
1\s
1!\\#\sdummy\scomment,\sduring\swait\n
	# execute # dummy comment, during wait
	# ELAPSED TIME
3l
1\s
2#	# "Makefile"
	# "RCS"
	# "README"
	# "dedrc"
	# "boxchars.c"
	# "ded.c"
	# "ded2s.c"
	# "dedblip.c"
	# "dedcmds.c"
	# "dedcolor.c"
	# "deddoit.c"
	# "deddump.c"
	# "dedfind.c"
	# "dedfree.c"
	# "dedline.c"
	# "dedmake.c"
	# "dedmsgs.c"
	# "dedname.c"
	# "dedread.c"
	# "dedring.c"
	# "dedscan.c"
	# "dedshow.c"
	# "dedsigs.c"
	# "dedsize.c"
	# "dedsort.c"
	# "dedtags.c"
	# "dedtype.c"
	# "deduniq.c"
	# "dedview.c"
	# "dedwait.c"
	# "dlog.c"
	# "ftree.c"
	# "history.c"
	# "inline.c"
	# "restat.c"
	# "showpath.c"
	# "sortset.c"
	# "ded.h"
	# "version.h"
1\s
1.	# execute # dummy comment, during wait
	# ELAPSED TIME
3l
1sN
2#
1V
1\s
1_
1sw
1.	# execute # dummy comment, during wait
	# ELAPSED TIME
3l
1ss
9b
1z
1\n
1k
1Z
1\s
1st
1Z
1V
1F	# chdir CM_TOOLS/test
1F	# chdir CM_TOOLS/bin
1F	# chdir CM_TOOLS/src
1F	# chdir CM_TOOLS/test
1F	# chdir CM_TOOLS/bin
1F	# chdir CM_TOOLS/src
1.	# execute # dummy comment, during wait
	# ELAPSED TIME
3l
1qyes\n
	# PROCESS ENDED
